%  SPM
% Written by Romina Kazemivala

clc;
clear all;
close all;
tic
%-------------------------------------------------------------------------
%%                          INPUT 
%-------------------------------------------------------------------------
%load Parameters
%CO, OUT, and RO Scattering Parameters
%1-3 of appendix C
load S11_CO.mat
load S11_RO.mat
load S11_OUT.mat
load S12_CO.mat
load S12_RO.mat
load S12_OUT.mat
load S21_CO.mat
load S21_RO.mat
load S21_OUT.mat
load S22_CO.mat
load S22_RO.mat
load S22_OUT.mat


%This uses all datasets
% s_CO = {S11_CO;S22_CO;S21_CO;S12_CO};
% s_RO = {S11_RO;S22_RO;S21_RO;S12_RO};
% s_OUT = {S11_OUT;S22_OUT;S21_OUT;S12_OUT};

%this uses only reflection datasets
s_CO = {S11_CO;S22_CO};
s_RO = {S11_RO;S22_RO};
s_OUT = {S11_OUT;S22_OUT};
z= 30e-3:10e-3:50e-3;

ftable = 3e9:1e9:8e9;
%-------------------------------------------------------------------------

%% Initialize Variables
%-------------------------------------------------------------------------


[nx, ny, nz,nf]=size(s_CO{1}); %nx, ny and nz are number of samples along x,y and z; fnum is number of frequency samples
ns = size(s_CO);
frng = 1:nf; %Change this if you want to see less data in the given bandwidth
COsize=10; %in mm
scan_step=10; %spatial sampling step in mm
c0 = 2.998e8; % Speed of light

% Relative permittivity of background medium
eps_RO = 1;
ROcontr=1;
% Relative permittivity of scattering probe in CO
eps_CO = 1.1;

% Volume of the scattering probe 
vol_sp = (10e-3)^3;

COsize=vol_sp; %in mm

ROcontr=eps_RO;
COcontr=eps_CO-eps_RO;

% Scan step
xstep = 10e-3;
xrng = xstep*(nx-1);

% Choose type of forward model: Born or Rytov
approx=1; % Born is 1, Rytov is 0. 

scaling_factor=(COsize^2)/(xstep^2); %calculating scaling factor for the cubic scattering probe. See Eq.(4.102)

%Activate frequency scaling
freq_scaling=1;


%Activate Apodization Filter
apod=0;
show_apod=0; %set to 1 for showing the apodization filter

%Activate Low-pass Filter
lpf=0;
show_lpf=0; %set to 1 for showing the LPF


if length(z)>1
    zstep = z(2)-z(1);
else
    zstep = z(1);
end
%-------------------------------------------------------------------------

%% Fill out s_RO if it contains only 1 value (i.e. size(s_RO{1} returns (1,1,nz,nf))
%-------------------------------------------------------------------------

if size(s_RO{1}(:,1,1,1)) == 1
    for s = 1:ns
        for f = 1:nf
        s_RO_new{s}(:,:,f) =  (squeeze(s_RO{s}(1,1,2,f))*ones(nx,ny));
        end
    end
end
s_RO = s_RO_new;

%--------------------------------------------------------------------------
%%                   Apodization Filter
%--------------------------------------------------------------------------
if apod

    tukeytol = 0.8;
    tukey = tukeywin(nx,tukeytol)*tukeywin(nx,tukeytol)';
    apod_filter = tukey;
else
    %if we dont want a filter, just multiply everything by 1!
    apod_filter = ones(nx);
end

 %show the apodization filter
if show_apod
    figure;
    imagesc(apod_filter);
    title('Apodization Filter (Tukey Filter)');
end

%==========================================================================
%% Obtain real domain datasets using either Born or Rytov's approximation
%==========================================================================
for s = 1:ns
    
    %Born's Approximation: 
    %Subtract RO from CO, OUT
    if approx
        %Extract  Scattered  Portion  of  OUT  Data
        s_OUT_scat{s} = s_OUT{s} - s_RO{s}; 
        for z = 1:nz
            %Extract  Scattered  Portion  of  PSF  Data
            s_CO_scat{s}(:,:,z,:) = squeeze(s_CO{s}(:,:,z,:)) - s_RO{s}; 
        end
    %Rytovs Approximation: 
    %Scattered Field OUT or CO = s_RO * ln (s_(OUT or CO)/s_RO)
    else
        %Extract  Scattered  Portion  of  OUT  Data
        s_OUT_scat{s} = s_RO{s}.*log(s_OUT{s}./s_RO{s});
        for z = 1:nz
                    %Extract  Scattered  Portion  of  PSF  Data
            s_CO_scat{s}(:,:,z,:) = s_RO{s}.*log(squeeze(s_CO{s}(:,:,:,z))./s_RO{s});
        end
    end
    
%==========================================================================    
                   %% Transform to Fourier domain
%==========================================================================    

    for f = frng
        
        
        if freq_scaling %If the frequency normalization is applied on the extracted data
        %prior to the Fourier transform
        maxi = max(max(max(abs(s_CO{s}(:,:,:,f)))));
        else
            maxi=1;
        end 
        s_OUT_fft{s}(:,:,f) = fft2(s_OUT_scat{s}(:,:,f).*apod_filter/maxi);

        for z = 1:nz
            s_CO_fft{s}(:,:,z,f) = fft2(s_CO_scat{s}(:,:,z,f).*apod_filter/maxi);

        end
    end
end
%--------------------------------------------------------------------------
%%                   Scattering Probe (SP) Power Maps
%--------------------------------------------------------------------------
 % line (9) in Image Reconstruction section of Appendix C

% The CO Maps need to be Nz by Nz at every kx, ky position. Structing the
% loops to achieve this

%First we sum across all S-parameters, performing the conjugate
%multiplication
M_CO_nonorm = zeros(nx,ny,nz,nz,nf);

%for CO we need to loop through z twice, once for the plane being measured (zp),
%and once for the PSF plane using to try and measure that particular plane
%(zn)
for zp = 1:nz
    for zn = 1:nz
        for f=frng
            % sum the conjugates across all S-parameters
           for s = 1:ns
              %See Equation (41) on Page 16 of the Report
               M_CO_nonorm(:,:,zp,zn,f) = M_CO_nonorm(:,:,zp,zn,f) + ...
                   s_CO_fft{s}(:,:,zn,f).*conj(s_CO_fft{s}(:,:,zp,f)); 

           end
        end
    end
end
M_CO = sum(M_CO_nonorm,5)/nf;


%--------------------------------------------------------------------------
%%                     Generate OUT Power Maps
%--------------------------------------------------------------------------

%The OUT Maps need to be Nz by 1 at every kx, ky position. Structing the
%loops to achieve this
M_OUT_nonorm = zeros(nx,ny,nz,nf);
M_OUT_norm = zeros(nx,ny,nz);

%OUT only has one plane being measured, so we loop through all PSFs only
for zn = 1:nz
    %we need to normalize at every frequency
    for f=frng
        % sum the conjugates across all S-parameters
        for s = 1:ns
           
            %See Equation (40) on Page 16 of the Report

            M_OUT_nonorm(:,:,zn,f) = M_OUT_nonorm(:,:,zn,f) + ...
                s_OUT_fft{s}(:,:,f).*conj(s_CO_fft{s}(:,:,zn,f));
            
        end
                M_OUT_norm(:,:,zn) = M_OUT_norm(:,:,zn)+ M_OUT_nonorm(:,:,zn,f);

    end
end

% After all the summation we divide by the number of frequencies
M_OUT = M_OUT_norm/nf;
%==========================================================================
%%                  Post-inversion Low-pass Filter
%==========================================================================

if lpf 
   
    %calculate minimum wavelength
    lambda_min = c0/sqrt(abs((eps_RO)))/ftable(nf);
    
    %Calculate the maximum angle the antenna can see. This assumes that the
    %directivity of the antenna does not affect the maximum angle, which is
    %typically not the case in experimental reconstructions
    z= 30e-3:10e-3:50e-3;

    alpha = atan(sqrt(2*(xrng/7)^2)/z(nz));
    
    %calculate maximum kmax, and determine is location in the physical
    %FFT images
    kmax = 4*pi*sin(alpha)/lambda_min;
    kx = pi/xstep;
    dkx = 2*kx/nx;
    kmax_steps = kmax/dkx;
    dev =kmax_steps;
    
    %set the order of the Butterworth filter
    order=4;
    bfilt = ones(nx,ny);
    %BUTTERWORTH FILTER
    for i = 1:nx
        for j =1:ny
            chg = ((i-((nx-1)/2+1))^2 + (j-((ny-1)/2+1))^2)^.5;
            % Create Butterworth filter.
            bfilt(i,j)= 1/(1 + (chg/dev)^(2*order));
        end
    end
    %Shift the filter to align with the FFT versions of the CO and OUT
    lpf = fftshift(bfilt);
    
    %Any values below 0.01 are automatically set to 0. This reduces
    %computational costs
    lpf = lpf.*(lpf>0.01);
    
        %TESTING: Make the LPF scale so that the overal summation of the filter
    %remains the same as if there was no window (i.e. try to move the
    %overal signal power from the erroneous high frequnecy responses to the
    %lower frequency values
    
    lpf = lpf/sum(sum(lpf))*nx*ny;
else
    %if we dont want to filter, just multiply everything by 1!
    lpf = ones(nx);
end


 %show the filter
if show_lpf
    figure;
    imagesc(ifftshift(lpf));
    title('Post-Inv. L.P. Filter');
end

%--------------------------------------------------------------------------
%%                       INVERSE PROBLEM
%--------------------------------------------------------------------------

%Solving a linear system of equations in Equation (42) on Page 16
contrast_func_fft = zeros(nx,ny,nz);
for kx=1:nx
    for ky=1:ny
        %Quantitative Contrast Reconstruction in Fourier Space Step 11 in
        %Apendix C
        contrast_func_fft(kx,ky,:)=pinv(squeeze(M_CO(kx,ky,:,:)))*squeeze(M_OUT(kx,ky,:)); 
    end
end

%Once we have the contrast function (in fourier domain) we can apply our
%multiplication factors and generate the permittivity map (eps_out)
eps_out=zeros(nx,ny,nz); 
for z=1:nz
    %Obtain  Real-space  Contrast  via  2D  Inverse  Fourier  Transform
    %Step 13 in Appendix C
    %Note that LPF is applied to the contrast function in the Fourier space
    eps_out(:,:,z)=scaling_factor.*COcontr.*ifftshift(ifft2(lpf.*contrast_func_fft(:,:,z)))+ROcontr; 
end

%--------------------------------------------------------------------------
%%                      Plotting the results 
%--------------------------------------------------------------------------
int=0; %interpolation yes 1, no 0. Note that the colormap values may change during certain interpolation settings.
int_rate=0.2; %set interpolation rate

if nz == 1 %plotting 2D images (this section can be also run to get the 3D images slice by slice separately: that is why loop over range locations is still there)
    
    for t=1:nz
        figure('Units', 'centimeters', 'OuterPosition',[0 10 70 14]);
        
        subplot(1, 3, 1) %qualitative outcome
        drawmap=ifftshift(ifft2(M_OUT(:,:,t)));
        
        V=abs(drawmap/max(max(drawmap))); %normalized modulus
        imagesc(0:(nx-1)*scan_step/10, 0:(ny-1)*scan_step/10,V);
        colorbar
        colormap parula
        shading interp
        alpha(0.9);
        
        subplot(1, 3, 2) %real part of the relative dielectric permittivity distribution
        drawmap=eps_out(:,:,t);
        if int == 1
            [X,Y]=meshgrid(1:nx,1:ny);
            [XI,YI]=meshgrid(1:int_rate:nx,1:int_rate:ny);
            V2=interp2(X,Y,drawmap,XI,YI,'spline');
            imagesc(0:(nx-1)*scan_step/10, 0:(ny-1)*scan_step/10,real(V2));
        else
            V=real(drawmap);
            imagesc(0:(nx-1)*scan_step/10, 0:(ny-1)*scan_step/10,V);
        end
        colorbar
        colormap parula
        shading interp
        alpha(0.9);
        
        subplot(1, 3, 3) %imaginary part of the relative dielectric permittivity distribution
        if int == 1
            [X,Y]=meshgrid(1:nx,1:ny);
            [XI,YI]=meshgrid(1:int_rate:nx,1:int_rate:ny);
            V2=interp2(X,Y,drawmap,XI,YI,'spline');
            imagesc(0:(nx-1)*scan_step/10, 0:(ny-1)*scan_step/10,imag(V2));
        else
            V=imag(drawmap);
            imagesc(0:(nx-1)*scan_step/10, 0:(ny-1)*scan_step/10,V);
        end
        colorbar
        colormap parula
        shading interp
        alpha(0.9);
        
        set(gcf, 'color','w');
    end
    
else % plotting 3D images
    
    if int == 1
        Z=ones(nx,ny);
        [X,Y]=meshgrid(1:nx,1:ny);
        [X1,Y1]=meshgrid(linspace(0,(nx-1)*scan_step/10,nx),linspace(0,(ny-1)*scan_step/10,ny));
        
        figure
        for k=1:nz
            [XI,YI]=meshgrid(1:int_rate:nx,1:int_rate:ny);
            V3(:,:,k)=interp2(X,Y,eps_out(:, :, k), XI,YI);
            surface(X1,Y1,Z.*k,real(V3(:,:,k)),'FaceColor','texturemap','CDataMapping','scaled')
        end
        colormap parula
        view(3), box on, axis tight square
        shading flat
        set(gca,'YDir','reverse','ZLim',[1 nz])
        colorbar
        
        figure
        for k=1:nz
            [XI,YI]=meshgrid(1:int_rate:nx,1:int_rate:ny);
            V3(:,:,k)=interp2(X,Y,eps_out(:, :, k), XI,YI);
            surface(X1,Y1,Z.*k,imag(V3(:,:,k)),'FaceColor','texturemap','CDataMapping','scaled')
        end
        colormap parula
        view(3), box on, axis tight square
        shading flat
        set(gca,'YDir','reverse','ZLim',[1 nz])
        colorbar
        
    else
        Z=ones(nx,ny);
        [X,Y]=meshgrid(1:nx,1:ny);
        figure
        for k=1:nz
            surface(X,Y,Z.*k,real(eps_out(:,:,k)),'FaceColor','texturemap','CDataMapping','scaled')
        end
        title('Real')
xlabel('x (cm)');
ylabel('y (cm)');
axis tight
view(30,40)
grid on
xh = get(gca,'XLabel'); % Handle of the x label
set(xh, 'Units', 'Normalized')
pos = get(xh, 'Position');
set(xh,'Rotation',25)
yh = get(gca,'YLabel'); % Handle of the y label
set(yh, 'Units', 'Normalized')
pos = get(yh, 'Position');
set(yh,'Rotation',-40)
        colormap parula
        view(3), box on, axis tight square
        shading flat
        set(gca,'YDir','reverse','ZLim',[1 nz])
        colorbar
        
        figure
        for k=1:nz
            surface(X,Y,Z.*k,imag(eps_out(:,:,k)),'FaceColor','texturemap','CDataMapping','scaled')
        end
        title('Imaginary')
        xlabel('x (cm)');
ylabel('y (cm)');
axis tight
view(30,40)
grid on
xh = get(gca,'XLabel'); % Handle of the x label
set(xh, 'Units', 'Normalized')
pos = get(xh, 'Position');
set(xh,'Rotation',25)
yh = get(gca,'YLabel'); % Handle of the y label
set(yh, 'Units', 'Normalized')
pos = get(yh, 'Position');
set(yh,'Rotation',-40)
        colormap parula
        view(3), box on, axis tight square
        shading flat
        set(gca,'YDir','reverse','ZLim',[1 nz])
        colorbar
        
    end
 end



toc
